"""Build click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import build
from yojenkins.cli import cli_build, cli_decorators
from yojenkins.cli.cli_utility import set_debug_log_level
from yojenkins.utility.utility import translate_kwargs


@build.command(short_help='\tBuild information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=bool, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def info(ctx, debug, **kwargs):
    """Build information"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.info(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tBuild status text/label')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number', metavar='INT')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=bool, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def status(ctx, debug, **kwargs):
    """Build status text/label"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.status(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())
    # FIXME: Showing running for job, seems not to have picked the right build ... when url to job is used!


@build.command(short_help='\tAbort build')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def abort(ctx, debug, **kwargs):
    """Abort build"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.abort(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tDelete build')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def delete(ctx, debug, **kwargs):
    """Delete build"""
    # TODO: Pass a list of build numbers
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.delete(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tGet build stages')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@cli_decorators.list
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def stages(ctx, debug, **kwargs):
    """Get build stages"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.stages(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tGet build logs')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.option('--tail', type=float, required=False, help='Last of logs. If < 1 then %, else number of lines')
@click.option('-dd',
              '--download-dir',
              type=click.Path(file_okay=False, dir_okay=True),
              required=False,
              is_flag=False,
              help='Download logs to directory')
@click.option('--follow',
              default=False,
              type=str,
              required=False,
              is_flag=True,
              help='Follow/Stream the logs as they are generated')
@click.pass_context
def logs(ctx, debug, **kwargs):
    """Get build logs"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.logs(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tOpen build in web browser')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def browser(ctx, debug, **kwargs):
    """Open build in web browser"""
    # TODO: Pass a list of build numbers
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.browser(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tStart monitor UI')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.option('-s', '--sound', type=bool, required=False, is_flag=True, help='Enable sound effects')
@click.pass_context
def monitor(ctx, debug, **kwargs):
    """Start monitor UI"""
    # TODO: Pass a list of build numbers
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.monitor(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tGet build parameters')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@cli_decorators.list
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def parameters(ctx, debug, **kwargs):
    """Get build parameters

    Build parameters are the parameters that were used to start the build.
    """
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.parameters(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())



"""Stage click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import stage
from yojenkins.cli import cli_decorators, cli_stage
from yojenkins.cli.cli_utility import set_debug_log_level
from yojenkins.utility.utility import translate_kwargs


@stage.command(short_help='\tStage information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.option('-j', '--job', type=str, required=False, help='Job name or URL')
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=bool, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def info(ctx, debug, **kwargs):
    """Stage information

    USAGE NOTES:

    \b
        - You must either sepcify the job with its build number
          or the direct URl to the build

    EXAMPLE:

    \b
    - yojenkins stage info my-build-stage --job my-job --latest
    - yojenkins stage info stage-name --build_url http://localhost:8080/job/my-job/3/ --pretty
    """
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_stage.info(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@stage.command(short_help='\tStage status text')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.option('-j', '--job', type=str, required=False, help='Job name or URL')
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=bool, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def status(ctx, debug, **kwargs):
    """Stage status text"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_stage.status(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@stage.command(short_help='\tGet stage steps')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@cli_decorators.list
@click.argument('name', nargs=1, type=str, required=True)
@click.option('-j', '--job', type=str, required=False, help='Job name or URL')
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=bool, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def steps(ctx, debug, **kwargs):
    """Get stage steps"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_stage.steps(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@stage.command(short_help='\tStage steps')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.option('-j', '--job', type=str, required=False, help='Job name or URL')
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Build URL (No job info needed)')
@click.option('--latest', type=bool, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.option('-dd',
              '--download_dir',
              type=click.Path(file_okay=False, dir_okay=True),
              required=False,
              is_flag=False,
              help='Download logs to directory')
@click.pass_context
def logs(ctx, debug, **kwargs):
    """Stage logs"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_stage.logs(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())



"""Importing sub-modules"""

from .account import Account
from .auth import Auth
from .credential import Credential
from .folder import Folder
from .jenkins_item_classes import JenkinsItemClasses
from .jenkins_item_config import JenkinsItemConfig
from .jenkins_item_template import JenkinsItemTemplate
from .job import Job
from .node import Node
from .rest import Rest
from .server import Server
from .stage import Stage
from .step import Step
from .yojenkins import YoJenkins



"""Node Menu CLI Entrypoints"""

import json
import logging

import click
import xmltodict

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history

# Getting the logger reference
logger = logging.getLogger()


@log_to_history
def info(profile: str, token: str, name: str, depth: int, **kwargs) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    data = yj_obj.node.info(name, depth)
    cu.standard_out(data, **kwargs)


@log_to_history
def list(profile: str, token: str, opt_list: bool, depth: int, **kwargs) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    data, data_list = yj_obj.node.list(depth)
    data = data_list if opt_list else data
    cu.standard_out(data, **kwargs)


@log_to_history
def create_permanent(profile: str, token: str, **kwargs) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    yj_obj.node.create_permanent(**kwargs)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def delete(profile: str, token: str, name: str) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    yj_obj.node.delete(name)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def disable(profile: str, token: str, name: str, message: str) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    yj_obj.node.disable(name, message)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def enable(profile: str, token: str, name: str, message: str) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    yj_obj.node.enable(name, message)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def config(profile: str, token: str, opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, opt_json: bool,
           name: str, filepath: str) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    data = yj_obj.node.config(filepath=filepath,
                              node_name=name,
                              opt_json=opt_json,
                              opt_yaml=opt_yaml,
                              opt_toml=opt_toml)
    opt_xml = not any([opt_json, opt_yaml, opt_toml])
    data = data if opt_xml else json.loads(json.dumps(xmltodict.parse(data)))
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def reconfig(profile: str, token: str, name: str, config_file: str, config_is_json: bool) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    yj_obj.node.reconfig(config_file=config_file, node_name=name, config_is_json=config_is_json)
    click.secho('success', fg='bright_green', bold=True)



"""Jenkins item template definition"""

from enum import Enum


class JenkinsItemTemplate(Enum):
    """Enum of Jenkins item templates ordered in categories

    Usage Examples:
        - `JenkinsItemTemplate.CREDENTIAL.value['ssh-user']`
    """

    CREDENTIAL = {
        "user-pass":
            '''
<com.cloudbees.plugins.credentials.impl.UsernamePasswordCredentialsImpl>
    <scope>${CRED_SCOPE_DOMAIN}</scope>
    <id>${CRED_ID}</id>
    <description>${CRED_DESCRIPTION}</description>
    <username>${CRED_USERNAME}</username>
    <password>${CRED_PASSWORD}</password>
    <usernameSecret>${CRED_USERNAME_AS_SECRET}</usernameSecret>
</com.cloudbees.plugins.credentials.impl.UsernamePasswordCredentialsImpl>''',
        "secret-text":
            '''
<org.jenkinsci.plugins.plaincredentials.impl.StringCredentialsImpl>
    <scope>${CRED_SCOPE_DOMAIN}</scope>
    <id>${CRED_ID}</id>
    <description>${CRED_DESCRIPTION}</description>
    <secret>${CRED_SECRET_TEXT}</secret>
</org.jenkinsci.plugins.plaincredentials.impl.StringCredentialsImpl>''',
        "ssh-key":
            '''
<com.cloudbees.jenkins.plugins.sshcredentials.impl.BasicSSHUserPrivateKey>
    <scope>${CRED_SCOPE_DOMAIN}</scope>
    <id>${CRED_ID}</id>
    <description>${CRED_DESCRIPTION}</description>
    <username>${CRED_USERNAME}</username>
    <usernameSecret>${CRED_USERNAME_AS_SECRET}</usernameSecret>
    <privateKeySource class="com.cloudbees.jenkins.plugins.sshcredentials.impl.BasicSSHUserPrivateKey$DirectEntryPrivateKeySource">
        <privateKey>${CRED_PRIVATE_KEY}</privateKey>
    </privateKeySource>
</com.cloudbees.jenkins.plugins.sshcredentials.impl.BasicSSHUserPrivateKey>'''
    }



import click

from yojenkins.__main__ import account
from yojenkins.cli import cli_account, cli_decorators
from yojenkins.cli.cli_utility import set_debug_log_level
from yojenkins.utility.utility import translate_kwargs


@account.command(short_help='\tList all users')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.list
@cli_decorators.profile
def list(debug, **kwargs):
    """List all users"""
    set_debug_log_level(debug)
    cli_account.list(**translate_kwargs(kwargs))


@account.command(short_help='\tGet user information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('user-id', nargs=1, type=str, required=True)
def info(debug, **kwargs):
    """Get user information"""
    set_debug_log_level(debug)
    cli_account.info(**translate_kwargs(kwargs))


@account.command(short_help='\tCreate a user account')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('user-id', nargs=1, type=str, required=True)
@click.argument('password', nargs=1, type=str, required=True)
@click.option('--is-admin',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='User has administrator control')
@click.option('--email', type=str, required=False, help='User email address')
@click.option('--description', type=str, required=False, help='User description')
def create(debug, **kwargs):
    """Create a user account"""
    set_debug_log_level(debug)
    cli_account.create(**translate_kwargs(kwargs))


@account.command(short_help='\tDelete a user account')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('user-id', nargs=1, type=str, required=True)
def delete(debug, **kwargs):
    """Delete a user account"""
    set_debug_log_level(debug)
    cli_account.delete(**translate_kwargs(kwargs))


@account.command(short_help='\tAdd or remove user permission')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('user_id', nargs=1, type=str, required=True)
@click.option('--action',
              type=click.Choice(['add', 'remove'], case_sensitive=False),
              required=True,
              help='Add or remove permission action')
@click.option('--permission-id', type=str, required=True, help='ID(s) of permission (comma separated)')
def permission(debug, **kwargs):
    """
    Add or remove user permission

    \b
    IMPORTANT:
      - Permission IDs must have correct capitalization or naming!
      - If permission is not found, read class docs, or try variation.

    \b
    EXAMPLES:
      - Will not work/match:  hudson.model.View.Create
      - Will work/match:      hudson.model.View.CREATE
    \b
      - Will not work/match:  hudson.security.Permission.GenericCreate
      - Will work/match:      hudson.security.Permission.CREATE
    """
    set_debug_log_level(debug)
    cli_account.permission(**translate_kwargs(kwargs))


@account.command(short_help='\tList all available permissions')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.list
@cli_decorators.profile
def permission_list(debug, **kwargs):
    """List all available permissions

    \b
    IMPORTANT:
      - Permission IDs may not have correct capitalization!
      - If permission is not found, read class docs, or try variation.

    \b
    EXAMPLE:
      - Not Correct:  hudson.model.View.Create
      - Correct:      hudson.model.View.CREATE
    \b
      - Will not work/match:  hudson.security.Permission.GenericCreate
      - Will work/match:      hudson.security.Permission.CREATE
    """
    set_debug_log_level(debug)
    cli_account.permission_list(**translate_kwargs(kwargs))


@account.command(short_help='\tReset a user password')
@cli_decorators.debug
@cli_decorators.profile
def password_reset(debug, **kwargs):
    """Reset a user password"""
    set_debug_log_level(debug)
    click.secho('TODO :-/', fg='yellow')



"""Account class definition"""

import logging
import os
from typing import Tuple

from yojenkins.utility import utility
from yojenkins.utility.utility import fail_out
from yojenkins.yo_jenkins.rest import Rest

# Getting the logger reference
logger = logging.getLogger()


class Account():
    """Account Class"""

    def __init__(self, rest: Rest) -> None:
        """Object constructor method, called at object creation

        Args:
            None
        """
        self.rest = rest
        self.groovy_script_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'groovy_scripts')

    def list(self) -> Tuple[list, list]:
        """List all accounts for the server

        Args:
            None

        Returns:
            List of credentials in dictionary format and a list of credential names
        """
        script_filepath = os.path.join(self.groovy_script_directory, 'user_list.groovy')
        account_list, success, error = utility.run_groovy_script(script_filepath=script_filepath,
                                                                 json_return=True,
                                                                 rest=self.rest)
        if not success:
            fail_out(f'Failed to list account. {error}')

        # Get a list of only account ids
        account_list_id = [account["id"] for account in account_list if "id" in account]

        logger.debug(f'Number of user accounts found: {len(account_list)}')
        logger.debug(f'Found the following user account ids: {account_list_id}')

        return account_list, account_list_id

    def info(self, user_id: str) -> dict:
        """Get information about an account

        Args:
            user_id: User account ID

        Returns:
            Dictionary of account information
        """
        script_filepath = os.path.join(self.groovy_script_directory, 'user_list.groovy')
        user_list, success, error = utility.run_groovy_script(script_filepath=script_filepath,
                                                              json_return=True,
                                                              rest=self.rest)
        if not success:
            fail_out(f'Failed to get account info. {error}')
        for user in user_list:
            if user['id'] == user_id:
                logger.debug(f'Successfully found account: {user_id}')
                return user
        fail_out(f'Failed to find account: {user_id}')

    def create(self, user_id: str, password: str, is_admin: bool, email: str, description: str) -> bool:
        """Create a new user account

        Args:
            user_id: Username
            password: Password
            is_admin: Is admin
            email: Email
            description: Description

        Returns:
            True if the account was created, False otherwise
        """
        kwargs = {
            'user_id': user_id,
            'password': password,
            'is_admin': 'true' if is_admin else 'false',
            'email': '' if not email else email,
            'description': '' if not description else description
        }
        script_filepath = os.path.join(self.groovy_script_directory, 'user_create.groovy')
        _, success, error = utility.run_groovy_script(script_filepath=script_filepath,
                                                      json_return=False,
                                                      rest=self.rest,
                                                      **kwargs)
        if not success:
            fail_out(f'Failed to create account. {error}')
        return True

    def delete(self, user_id: str) -> bool:
        """Delete a user account

        Args:
            user_id: Username of account to be deleted

        Returns:
            True if the account was deleted, False otherwise
        """
        kwargs = {'user_id': user_id}
        script_filepath = os.path.join(self.groovy_script_directory, 'user_delete.groovy')
        _, success, error = utility.run_groovy_script(script_filepath=script_filepath,
                                                      json_return=False,
                                                      rest=self.rest,
                                                      **kwargs)
        if not success:
            fail_out(f'Failed to delete account. {error}')
        return True

    def permission(self, user_id: str, action: str, permission_id: str) -> bool:
        """Add or remove user account permissions

        Details:
            The permission name can be a comma separated list of permission names

        Args:
            user_id: Username of account to be deleted
            action: Action to perform, either 'add' or 'remove'
            permission_id: Permission name to add or remove (can be a comma separated list)

        Returns:
            True if the permissions were added, False otherwise
        """
        # Parse comma seperated string
        permission_list = utility.parse_and_check_input_string_list(permission_id, join_back_char=', ')
        permission_groovy_list = "[" + permission_list + "]"

        if action == 'add':
            logger.debug(f'Adding the following permissions to user "{user_id}": {permission_list}')
            kwargs = {
                'user_id': user_id,
                'permission_groovy_list': permission_groovy_list,
                'permission_enabled': 'true'
            }
        elif action == 'remove':
            logger.debug(f'Removing the following permissions from user "{user_id}": {permission_list}')
            kwargs = {
                'user_id': user_id,
                'permission_groovy_list': permission_groovy_list,
                'permission_enabled': 'false'
            }
        else:
            fail_out(f'Invalid permission action specified: {action}')

        script_filepath = os.path.join(self.groovy_script_directory, 'user_permission_add_remove.groovy')
        _, success, error = utility.run_groovy_script(script_filepath=script_filepath,
                                                      json_return=False,
                                                      rest=self.rest,
                                                      **kwargs)
        if not success:
            fail_out(f'Failed to {action} account permissions. {error}')
        return True

    def permission_list(self) -> Tuple[list, list]:
        """Get all the available permissions and descriptions

        Args:
            None

        Returns:
            Dictionary of availabe permissions and descriptions
        """
        script_filepath = os.path.join(self.groovy_script_directory, 'user_permission_list.groovy')
        permission_list, success, error = utility.run_groovy_script(script_filepath=script_filepath,
                                                                    json_return=True,
                                                                    rest=self.rest)
        if not success:
            fail_out(f'Failed to list all available permissions. {error}')

        # Capitalize last part of permission ID and remove "GENERIC" sub-string
        # Example: "hudson.security.Permission.GenericRead" becomes "hudson.security.Permission.READ"
        for permission in permission_list:
            if 'id' in permission:
                items = permission['id'].split('.')
                items[-1] = items[-1].upper().replace("GENERIC", "")
                permission['id'] = '.'.join(items)

        # Get a list of only permission ids
        permission_list_ids = [permission["id"] for permission in permission_list if "id" in permission]

        logger.debug(f'Number of permission found: {len(permission_list)}')
        logger.debug(f'Found the following permission ids: {permission_list_ids}')

        return permission_list, permission_list_ids



"""Auth class definition"""

import json
import logging
import os
import re
import sys
from datetime import datetime
from getpass import getpass
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, Tuple

import toml
from jenkins import Jenkins as JenkinsSDK

from yojenkins.utility import utility
from yojenkins.utility.utility import TextStyle, fail_out, failures_out, print2
from yojenkins.yo_jenkins.rest import Rest

# Getting the logger reference
logger = logging.getLogger()

# TODO: Find centralized location for these static values
CONFIG_DIR_NAME = '.yojenkins'
CREDS_FILE_NAME = 'credentials'
PROFILE_ENV_VAR = 'YOJENKINS_PROFILE'

REQUIRED_PROFILE_KEYS = ['jenkins_server_url', 'username']
ALLOWED_PROFILE_KEYS = ['jenkins_server_url', 'username', 'api_token', 'active']


class Auth:
    """Handeling of authentication and profile management functionality"""

    def __init__(self, rest=None) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        # Request object
        if not rest:
            # If not provided, make a new object
            self.rest: object = Rest()
        else:
            self.rest: object = rest

        # JenkinsSDK object - Instantiated in create_auth()
        self.jenkins_sdk = None

        self.jenkins_profile: Dict[str, Any] = {}
        self.jenkins_username = ''
        self.jenkins_api_token = ''
        self.authenticated = False

    def _update_profiles(self, profiles: dict) -> bool:
        """Create/Update the current credentials profile file

        Details: This method will overwrite any previous file content.
                 It will also add a file prefix for TOML formats

        Args:
            profiles : All profile info to be written to file

        Returns:
            True if successfull, else False
        """
        # Storing configurations in file
        output_path = os.path.join(Path.home(), CONFIG_DIR_NAME, CREDS_FILE_NAME)
        logger.debug(f'Saving new file (TOML format): "{output_path}" ...')
        with open(os.path.join(output_path), 'w') as file:  # Overwrite previous content
            toml.dump(profiles, file)

        # Add top file prefix for TOML file format
        lines_new = [f'# -*- mode: toml -*-{os.linesep}', f'# vim: set filetype=toml{os.linesep}']
        success = utility.append_lines_to_file(output_path, lines_new, 'beginning')
        if not success:
            return False

        return True

    def get_rest(self) -> object:
        """Convenience method to return the REST object

        Args:
            None

        Returns:
            Rest Object
        """
        return self.rest

    def generate_token(self,
                       token_name: str = '',
                       server_base_url: str = '',
                       username: str = '',
                       password: str = '') -> str:
        """Generate a Jenkins server API token with the specified server/username

        Args:
            token_name      : (Optional) Name of new API token. Will prompt if not passed.
            server_base_url : (Optional) Server base URL. Will prompt if not passed.
            username        : (Optional) Username. Will prompt if not passed.
            password        : (Optional) Password. Will prompt if not passed.

        Returns:
            Server API Token
        """
        # Ask for anything not passed
        if not token_name:
            prompt_text = TextStyle.BOLD + TextStyle.YELLOW + "Enter desired API TOKEN NAME: " + TextStyle.NORMAL
            token_name = input(prompt_text)
            logger.debug(f'User input: {token_name}')

        if not server_base_url:
            prompt_text = TextStyle.BOLD + TextStyle.YELLOW + "Enter Jenkins SERVER BASE URL: " + TextStyle.NORMAL
            server_base_url = input(prompt_text)
            logger.debug(f'User input: {server_base_url}')

        if not username:
            prompt_text = TextStyle.BOLD + TextStyle.YELLOW + 'Enter Jenkins server USERNAME: ' + TextStyle.NORMAL
            username = input(prompt_text)
            logger.debug(f'User input: {username}')

        if not password:
            prompt_text = TextStyle.BOLD + TextStyle.YELLOW + f'Enter "{username}" PASSWORD: ' + TextStyle.NORMAL
            password = getpass(prompt=prompt_text, stream=None)
            logger.debug('User entered password')

        # Requesting the Jenkins crumb
        request_return_text, _, success = self.rest.request(target=server_base_url.strip('/') + '/crumbIssuer/api/xml',
                                                            is_endpoint=False,
                                                            request_type='get',
                                                            json_content=False,
                                                            auth=(username, password),
                                                            new_session=True)
        if not success:
            fail_out('Failed to generate API token. Crumb issuer failed. '
                     'Check server base URL, username, or password.')
        crumb_value = re.sub(re.compile('<.*?>|Jenkins-Crumb'), '', request_return_text)
        logger.debug(f'Request crumb value: {crumb_value}')

        # Generating the token
        headers = {'Jenkins-Crumb': crumb_value}
        params = (
            ('newTokenName', token_name),
            ('tree', 'data[tokenValue]'),
        )
        request_return_content, _, success = self.rest.request(
            target=server_base_url.strip('/') +
            '/me/descriptorByName/jenkins.security.ApiTokenProperty/generateNewToken/api/json',
            is_endpoint=False,
            request_type='post',
            json_content=True,
            auth=(username, password),
            new_session=False,
            params=params,
            headers=headers)
        if not success:
            fail_out('Failed to generate API token. Request failed')

        generated_token = ''
        try:
            generated_token = request_return_content['data']['tokenValue']
            logger.debug(f'Successfully generated server API token "{token_name}"!')
        except KeyError:
            fail_out('Failed to generate API token. Failed to find "tokenValue" in the POST request return header')

        return generated_token

    def profile_add_new_token(self, profile_name: str) -> str:
        """Update the specified credentials profile with a new server API token

        Args:
            profile_name : Name of the profile as specified in the credentials file

        Returns:
            Server API Token
        """
        # Check if the credential config file exists
        file_exists, file_path = self._detect_creds_file()
        if not file_exists:
            fail_out('Credential configuration ')

        # Load the current cred config file
        profiles = utility.load_contents_from_local_file("toml", file_path)
        if not profiles:
            return ''

        # Check if the profile exists
        if profile_name not in profiles:
            fail_out(f'Profile "{profile_name}" does not exist in the credentials file')

        profile_info = profiles[profile_name]
        logger.debug(f'Profile "{profile_name}" loaded')
        logger.debug(f'Profile info: {profile_info}')

        # Generate a API token name (seen in Jenkins UI)
        token_name = f'yojenkins_{datetime.now().strftime("%m-%d-%Y_%I-%M-%S")}'

        # Check if needed information is listed in the profile
        if 'jenkins_server_url' not in profile_info:
            logger.debug('No "jenkins_server_url" key listed in profile')
            server_url = ''
        else:
            server_url = profile_info['jenkins_server_url']
        if 'username' not in profile_info:
            logger.debug('No "username" key listed in profile')
            username = ''
        else:
            username = profile_info['username']

        # Generate and store the API token
        api_token = self.generate_token(token_name=token_name, server_base_url=server_url, username=username)
        profiles[profile_name]['api_token'] = api_token

        success = self._update_profiles(profiles=profiles)
        if not success:
            fail_out(f'Failed to add new API token to profile "{profile_name}"')

        return api_token

    def configure(self, auth_file: str = '') -> bool:
        """Configure/add a new credentials profile

        Details: This method will prompt the user with a series of questions
                 to add a new profile to the credentials profile file

        Args:
            auth_file : (Optional) Path to the the authentication setup JSON file

        Returns:
            This is a description of what is returned.
        """
        config_dir_abs_path = os.path.join(Path.home(), CONFIG_DIR_NAME)
        creds_file_abs_path = os.path.join(config_dir_abs_path, CREDS_FILE_NAME)

        # Checking if credential config file exists
        file_exists, file_path = self._detect_creds_file()
        if file_exists:
            logger.debug(f'Credentials file found in current user home directory: {file_path}')
            logger.debug(f'Loading credentials file: {file_path} ...')

            # Load the current cred config file
            profiles = utility.load_contents_from_local_file("toml", file_path)
            logger.debug(f'Currently listed profile names: {", ".join(list(profiles.keys()))}')

            logger.debug(f'Credentials profile file found in current user home directory: {file_path}')
        else:
            logger.debug('Credentials file NOT found in current user configuration directory')
            logger.debug(f'Creating credentials file: {creds_file_abs_path} ...')

            # Create new profile from scratch
            profiles = {}

            logger.debug(f'Credentials profile file NOT found: "{creds_file_abs_path}"')
            logger.debug(f'Creating a new credentials file ...')

        # Check if user passed authentication info file
        if auth_file:
            # Load the authentication setup JSON file
            setup_info = utility.load_contents_from_local_file("json", auth_file)
            if not setup_info:
                fail_out(f'Failed to load authentication setup JSON info file: {auth_file}')

            print2(f'Adding {len(setup_info)} authentication profile(s) specified in file "{auth_file}" ...')

            # Check each provided authentication setup profile
            # FIXME: Implement JSON Schema validation !!!!
            for index, (setup_profile_name, setup_profile_info) in enumerate(setup_info.items()):
                print2('')
                print2(f'[{index + 1} of {len(setup_info)}] Adding profile: {setup_profile_name} ...')

                failed_to_add = False

                # Check allowed keys
                keys_to_remove = []
                for key in setup_profile_info.keys():
                    if key not in ALLOWED_PROFILE_KEYS:
                        logger.debug(f'  - Invalid key:  {key} - IGNORING')
                        keys_to_remove.append(key)
                [setup_profile_info.pop(key, None) for key in keys_to_remove]

                # Check required keys to setup a new profile
                for required_key in REQUIRED_PROFILE_KEYS:
                    if required_key not in setup_profile_info.keys():
                        print2(f'  - Required key: {required_key} - NOT FOUND', color='red')
                        print2(f'Failed to add authentication profile: {setup_profile_name}', color='red')
                        failed_to_add = True
                    else:
                        logger.debug(f'  - Required key: {required_key} - FOUND')

                        # Check if required key value is not empty
                        if not setup_profile_info[required_key]:
                            print2(f'  - Required key: {required_key} - EMPTY VALUE', color='red')
                            failed_to_add = True
                        else:
                            logger.debug(f'  - Required key: {required_key} - HAS VALUE')

                if failed_to_add:
                    print2(f'Failed to add authentication profile: {setup_profile_name}', color='red')
                    continue

                # Check if the profile already exists
                if setup_profile_name in profiles:
                    print2(f'  - Profile "{setup_profile_name}" already exists in the credentials file. Replacing ...',
                           color="yellow")

                # If API token is not there, add a blank for it
                if 'api_token' not in setup_profile_info:
                    setup_profile_info['api_token'] = ''

                # Set to active if not specified
                if 'active' not in setup_profile_info:
                    setup_profile_info['active'] = True

                # Add to profiles to add
                profiles[setup_profile_name] = setup_profile_info
        else:
            # Prompting user for details
            print2('Please enter the following information to create your first profile entry')
            print2(f'This information will be stored in "{creds_file_abs_path}"')
            print2('')
            profile_name = input(TextStyle.BOLD + TextStyle.YELLOW + '[ OPTIONAL ] Enter PROFILE NAME (default):  ' +
                                 TextStyle.NORMAL)
            profile_name = 'default' if not profile_name else profile_name
            if profile_name in profiles:
                print2('')
                print2(f'WARNING : You are about to overwrite the current profile "{profile_name}"')
                print2('')
            profiles[profile_name] = {}
            profiles[profile_name]['jenkins_server_url'] = input(TextStyle.BOLD + TextStyle.YELLOW +
                                                                 '[ REQUIRED ] Enter Jenkins SERVER BASE URL:  ' +
                                                                 TextStyle.NORMAL)
            profiles[profile_name]['username'] = input(TextStyle.BOLD + TextStyle.YELLOW +
                                                       '[ REQUIRED ] Enter USERNAME:  ' + TextStyle.NORMAL)

            # If set up manually, set as active by default
            profiles[profile_name]['active'] = True

        return self._update_profiles(profiles=profiles)

    def _detect_config_dir(self) -> Tuple[bool, str]:
        """Detect/find the configuration directory in the home directory

        Args:
            None

        Returns:
            Success, Directory path of the credentials profile file
        """
        home_dir = Path.home()
        if not os.path.exists(home_dir):
            logger.debug(f'Home directory does not exist: {home_dir}')
            return False, ''
        config_dir_abs_path = os.path.join(home_dir, CONFIG_DIR_NAME)
        if not os.path.exists(config_dir_abs_path):
            logger.debug(f'Configuration directory does not exist: {config_dir_abs_path}')
            return False, ''
        return True, config_dir_abs_path

    def _detect_creds_file(self) -> Tuple[bool, str]:
        """Detect/find the credentials profile file in the home directory

        Args:
            None

        Returns:
            Success, File path of the credentials profile file
        """
        if not self._detect_config_dir()[0]:
            return False, ''
        config_dir_abs_path = os.path.join(Path.home(), CONFIG_DIR_NAME)

        # Seeing if configuration file is in specified directories
        if os.path.exists(os.path.join(config_dir_abs_path, CREDS_FILE_NAME)):
            logger.debug(f'Successfully found credential file "{CREDS_FILE_NAME}" found in user '
                         f'configuration directory: {config_dir_abs_path}')
            config_filepath = os.path.join(config_dir_abs_path, CREDS_FILE_NAME)
            return True, config_filepath
        else:
            logger.debug(f'Failed to find credential file "{CREDS_FILE_NAME}" in user '
                         f'configuration directory: {config_dir_abs_path}')
            return False, ''

    def get_credentials(self, profile: str = '') -> Dict:
        """Get the contents of the credentials profiles file.

        Details: The order (preference) of specified profile credentials to get are the following:
            - Passed `profile` argument
            - Environmental Variable as specified in `PROFILE_ENV_VAR`
            - Profile listed as `default` in the credentials profile file
            - Any other / first profile set to active

        Args:
            profile : (Optional) Only get the configuration for particular profile

        Returns:
            The credential information of the specified credentials profile
        """
        profile_is_option_text = False
        try:
            # Load profile from CLI option value
            json.loads(profile)
            logger.debug('Passed profile is a JSON format string via --profile')
            profile_items_all = utility.load_contents_from_string('json', profile)
            profile_is_option_text = True
        except (JSONDecodeError, TypeError):
            # Check to see if user meant to provide entire profile
            # For security reasons do not show what user intendent to provide
            if profile:
                if profile.startswith('{'):
                    fail_out('Specified profile option is not valid. Its value begins with '
                             'a "{" symbol indicating user attempted to provide profile info '
                             'via --profile option. Please check option input structure.')
            creds_file_abs_path = os.path.join(Path.home(), CONFIG_DIR_NAME, CREDS_FILE_NAME)

            # Check if credentials file existis on system
            if not self._detect_creds_file()[0]:
                # If no configuration file found, configure one
                logger.debug('No credentials file found. Configuring one ...')
                self.configure()

            # Loading configurations file
            creds_info = utility.load_contents_from_local_file("toml", creds_file_abs_path)
            if not creds_info:
                fail_out(f'Failed to load credentials file: {creds_file_abs_path}')

            # Get the listed profiles
            profile_items_all = creds_info
            logger.debug(f'Number of listed profiles found: {len(profile_items_all)}')

        # TODO: Add JSON Schema validation for profile structure

        # Filter out the profiles that are misconfigured (missing keys)
        profile_items = {}
        logger.debug(f'Ignoring profiles that do not have at least the '
                     f'following keys: {", ".join(REQUIRED_PROFILE_KEYS)} ...')
        if profile_is_option_text:
            if all(item in list(profile_items_all.keys()) for item in REQUIRED_PROFILE_KEYS):
                profile_items["CLI_PROVIDED"] = profile_items_all
            else:
                logger.debug("Profile JSON object provided with text via --profile does not "
                             f'contain the required information keys: {", ".join(REQUIRED_PROFILE_KEYS)}')
        else:
            for i, (profile_name, profile_values) in enumerate(profile_items_all.items()):
                if not isinstance(profile_values, dict):
                    fail_out('Failed to find specified profile as the correct '
                             'key-value structure. Please check profile structure')
                if all(item in list(profile_values.keys()) for item in REQUIRED_PROFILE_KEYS):
                    logger.debug(f'    - Profile {i+1} of {len(profile_items_all)}: "{profile_name}" - OK')
                    profile_items[profile_name] = profile_values
                else:
                    logger.debug(f'    - Profile {i+1} of {len(profile_items_all)}: "{profile_name}" - IGNORED')
        if not profile_items:
            failures_out([
                'Failed to find any valid profiles in the provided credentials',
                f'A valid profile must have at least the follwing keys: {", ".join(REQUIRED_PROFILE_KEYS)}'
            ])

        # Select the credential profile
        profile_selected = {}

        # PRIORITY 1 - Argument --profile
        if profile:
            if profile_is_option_text:
                profile = list(profile_items.keys())[0]
                profile_selected = profile_items[profile]
                profile_selected['profile'] = profile
                logger.debug('Successfully loaded entire profile provided through --profile')
            else:
                if profile in profile_items:
                    profile_selected = profile_items[profile]
                    profile_selected['profile'] = profile
                    logger.debug(f'Successfully matched specified --profile "{profile}"')
                else:
                    fail_out(f'Failed to find the profile "{profile}" within all loaded '
                             f'profiles: {", ".join(list(profile_items.keys()))}')
        else:
            logger.debug('Argument "--profile" was not specified')

        # PRIORITY 2 - Environmental Variable
        if not profile_selected:
            if PROFILE_ENV_VAR in os.environ:
                env_var_value = os.getenv(PROFILE_ENV_VAR)
                profile = env_var_value if env_var_value else ''
                if profile in profile_items:
                    profile_selected = profile_items[profile]
                    profile_selected['profile'] = profile
                    logger.debug(
                        f'Successfully matched set {PROFILE_ENV_VAR} environmental variable value "{profile}"')
                else:
                    fail_out(f'Failed to find the set environmental variable {PROFILE_ENV_VAR} "{profile}" '
                             'in the profiles loaded')
            else:
                logger.debug(f'Environmental Variable "{PROFILE_ENV_VAR}" not set')

        # PRIORITY 3 - "default" profile
        if not profile_selected:
            profile = 'default'
            if profile in profile_items:
                profile_selected = profile_items[profile]
                profile_selected['profile'] = profile
                logger.debug('Successfully found "default" profile in the configuration file')

            else:
                logger.debug('Default profile "default" not found')

        # PRIORITY 4 - Any other active one
        if not profile_selected:
            logger.debug('Selecting the first listed active profile ...')
            for profile, profile_values in profile_items.items():
                if 'active' in profile_values:
                    if profile_values['active'] == True:
                        profile_selected = profile_items[profile]
                        profile_selected['profile'] = profile
                        logger.debug(f'Successfully selected active profile: "{profile}"')
                        break

        if not profile_selected:
            fail_out('No active user profiles found. If no profile was set up, please configure one')

        # Show the partial info of the loaded profile
        self.jenkins_profile = profile_selected

        logger.debug('The following profile has been loaded:')
        logger.debug(f'    - Profile:             {self.jenkins_profile["profile"]}')
        logger.debug(f'    - Jenkins Server URL:  {self.jenkins_profile["jenkins_server_url"]}')
        logger.debug(f'    - Username:            {self.jenkins_profile["username"]}')

        return profile_selected

    def create_auth(self, profile_info: dict = {}, token: str = '') -> bool:
        """Authenticate with the Jenkins server

        Details: If not server API token in the profile used, ask for token/password

        Args:
            profile_info : (Optional) Credentials profile information
            token:         API token as passed to override profile value

        Returns:
            True if successfully authenticated, else False
        """
        # If the profile information is passed, use it
        if profile_info:
            self.jenkins_profile = profile_info
        if not self.jenkins_profile:
            fail_out('No credential profile loaded')

        # Check if server url has a protocol schema
        url_protocol_schema = re.findall('(\w+)://', self.jenkins_profile["jenkins_server_url"])
        if not url_protocol_schema:
            fail_out('Failed to find a valid server URL protocol schema (ie. http://, https://, etc) '
                     f'in the loaded profile server url: "{self.jenkins_profile["jenkins_server_url"]}"')

        # Check if password is listed, if not, ask for it
        if 'api_token' not in self.jenkins_profile or not self.jenkins_profile['api_token']:
            print2('')
            prompt_text = f'Profile {self.jenkins_profile["profile"]} does not contain a "api_token" key'
            print2(prompt_text, bold=True, color='yellow')
            prompt_text = TextStyle.BOLD + TextStyle.YELLOW + f"Enter Jenkins password or server API Token for user {self.jenkins_profile['username']}: " + TextStyle.NORMAL
            self.jenkins_profile['api_token'] = getpass(
                prompt=prompt_text,
                stream=None,
            )
            print2('')

            # NOTE: Below dead code is to potentially check length of the api_token
            # if len(self.jenkins_profile['api_token']) < 5:
            #     logger.debug(f'The entered API Token has a length of {len(self.jenkins_profile["api_token"])}, which is too short')
            #     return False

        # Load the token or password
        if token:
            #  logger.debug('API Token / Password loaded via --token CLI option')
            self.jenkins_profile['api_token'] = token
            self.jenkins_api_token = token
            hidden_token = "Loaded via --token"
        else:
            self.jenkins_api_token = self.jenkins_profile['api_token']
            hidden_token = ''.join(["*" for _ in self.jenkins_profile['api_token']])
        logger.debug(f'    - API Token:           {hidden_token}')
        self.jenkins_username = self.jenkins_profile['username']

        # Creating Jenkins SDK object(Exception handling: jenkins.JenkinsException)
        try:
            self.jenkins_sdk = JenkinsSDK(url=self.jenkins_profile['jenkins_server_url'],
                                          username=self.jenkins_profile['username'],
                                          password=self.jenkins_profile['api_token'],
                                          timeout=10)
        except Exception as error:
            fail_out(f'Internal Error: Failed to create Jenkins object. Exception: {error}')

        # Update the credentials in Rest object
        self.rest.set_credentials(username=self.jenkins_profile['username'],
                                  api_token=self.jenkins_profile['api_token'],
                                  server_url=self.jenkins_profile['jenkins_server_url'])

        # Check network connection
        if not self.rest.is_reachable():
            print2(f'Jenkins server connection failed (Server: {self.jenkins_profile["jenkins_server_url"]})',
                   bold=True,
                   color='red')
            print2('Possible causes:', bold=True, color='red')
            print2(f'  - Wrong Jenkins server URL: {self.jenkins_profile["jenkins_server_url"]}',
                   bold=True,
                   color='red')
            print2('  - Network/Internet is down', bold=True, color='red')
            print2('  - Server container is down', bold=True, color='red')
            print2('Possible solutions:', bold=True, color='red')
            print2('   - Fix yo network connection to server', bold=True, color='red')
            print2('   - Check if the server container or container engine is up and running', bold=True, color='red')
            sys.exit(1)

        # Checking authentication
        logger.debug(f'Checking authentication to Jenkins server: {self.jenkins_profile["jenkins_server_url"]} ...')
        if not self.verify():
            # TODO: Move this message to cli_auth.py, only return bool
            print2(f'Jenkins server authentication failed (Username: {self.jenkins_profile["username"]})',
                   bold=True,
                   color='red')
            print2('Possible causes:', bold=True, color='red')
            print2(f'    - Wrong Jenkins server URL: {self.jenkins_profile["jenkins_server_url"]}',
                   bold=True,
                   color='red')
            print2(f'    - Incorrect username: {self.jenkins_profile["username"]}', bold=True, color='red')
            print2('    - Incorrect, removed, or expired API Token', bold=True, color='red')
            print2(f'    - Username, {self.jenkins_profile["username"]}, does not have permission',
                   bold=True,
                   color='red')
            print2('    - Jenkins server is still in the process of starting up', bold=True, color='red')
            print2('Possible solutions:', bold=True, color='red')
            print2('    - yojenkins auth token', bold=True, color='red')
            print2('    - yojenkins auth configure', bold=True, color='red')
            print2('    - Manually create or update credentials file in home directory', bold=True, color='red')
            print2('    - Go to Jenkins Web UI and check user configurations', bold=True, color='red')
            print2('    - Give Jenkins server a little to start up', bold=True, color='red')
            sys.exit(1)

        return True

    def show_local_credentials(self) -> Dict:
        """Output/Display the credentials profile file

        Args:
            None

        Returns:
            Contents of the credentials profile file
        """
        success, config_filepath = self._detect_creds_file()
        if not success:
            # If no configuration file found
            fail_out(f'Failed to find a valid credentials file')

        # Loading configurations file
        return utility.load_contents_from_local_file("toml", config_filepath)

    def verify(self) -> bool:
        """Verify/Check if current credentials can authenticate with server

        Args:
            None

        Returns:
            True if successfully authenticated, else False
        """
        logger.debug('Verifying server authentication by requesting user information ...')
        request_url = ''
        try:
            request_url = self.jenkins_profile['jenkins_server_url'].strip('/') + "/me/api/json"
        except KeyError:
            fail_out('Failed to find profile key "jenkins_server_url". Profile may not have loaded correctly')
        request_success = self.rest.request(target=request_url, is_endpoint=False, request_type='head')[2]
        if not request_success:
            fail_out('Failed server authentication with specified user credentials')
        logger.debug('Successfully authenticated with specified user credentials')

        return True

    def user(self) -> Dict:
        """Get current user information

        Details: Targeting the user that is specified in the selected profile

        Args:
            None

        Returns:
            User information
        """
        data = self.rest.request('me/api/json', 'get')[0]
        if not data:
            fail_out('Failed to get user information')
        return data



"""Job click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import job
from yojenkins.cli import cli_decorators, cli_job
from yojenkins.cli.cli_utility import set_debug_log_level
from yojenkins.utility.utility import translate_kwargs


@job.command(short_help='\tJob information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
def info(debug, **kwargs):
    """Job information"""
    set_debug_log_level(debug)
    cli_job.info(**translate_kwargs(kwargs))


@job.command(short_help='\tSearch jobs by REGEX pattern')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('search_pattern', nargs=1, type=str, required=True)
@click.option('-sf', '--search-folder', type=str, default='', required=False, help='Folder within which to search')
@click.option('-d', '--depth', type=int, default=4, required=False, help='Search depth from root directory')
@click.option('-fn',
              '--fullname',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Search entire job path name')
@cli_decorators.list
def search(debug, **kwargs):
    """Search jobs by REGEX pattern"""
    set_debug_log_level(debug)
    cli_job.search(**translate_kwargs(kwargs))


@job.command(short_help='\tList all builds for job')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
@cli_decorators.list
def list(debug, **kwargs):
    """List all builds for job"""
    set_debug_log_level(debug)
    cli_job.build_list(**translate_kwargs(kwargs))


@job.command(short_help='\tGet next build number')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
def next(debug, **kwargs):
    """Get next build number"""
    set_debug_log_level(debug)
    cli_job.build_next(**translate_kwargs(kwargs))


@job.command(short_help='\tGet previous build number')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
def last(debug, **kwargs):
    """Get previous build number"""
    set_debug_log_level(debug)
    cli_job.build_last(**translate_kwargs(kwargs))


@job.command(short_help='\tSet the next build number')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
@click.argument('build_number', nargs=1, type=int, required=True)
def set(debug, **kwargs):
    """Set the next build number"""
    set_debug_log_level(debug)
    cli_job.build_set(**translate_kwargs(kwargs))


@job.command(short_help='\tCheck if build number exists')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
@click.argument('build_number', nargs=1, type=int, required=True)
def build_exist(debug, **kwargs):
    """Check if build number exists"""
    set_debug_log_level(debug)
    cli_job.build_exist(**translate_kwargs(kwargs))


@job.command(short_help='\tBuild a job')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
@click.option('-p',
              '--parameter',
              nargs=2,
              multiple=True,
              required=False,
              help='Specify key-value parameter. Can use multiple times. Use once per parameter')
def build(debug, **kwargs):
    """Build a job"""
    set_debug_log_level(debug)
    cli_job.build(**translate_kwargs(kwargs))


@job.command(short_help='\tCheck if this job is in queue')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
@click.option('-i', '--id', type=bool, default=False, required=False, is_flag=True, help='Only output queue ID')
def queue_check(debug, **kwargs):
    """Check if this job is in queue"""
    set_debug_log_level(debug)
    cli_job.queue_check(**translate_kwargs(kwargs))


@job.command(short_help='\tCancel this job in queue')
@cli_decorators.debug
@cli_decorators.profile
@click.option('-i', '--id', type=int, default=False, required=True, help='Queue ID')
def queue_cancel(debug, **kwargs):
    """Cancel this job in queue"""
    set_debug_log_level(debug)
    # cli_job.queue_cancel(**translate_kwargs(kwargs))
    click.secho('TODO :-/', fg='yellow')


@job.command(short_help='\tOpen job in web browser')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
def browser(debug, **kwargs):
    """Open job in web browser"""
    set_debug_log_level(debug)
    cli_job.browser(**translate_kwargs(kwargs))


@job.command(short_help='\tGet job configuration')
@cli_decorators.debug
@cli_decorators.format_output
@click.option('-j',
              '--json',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Output config in JSON format')
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
@click.option('--filepath',
              type=click.Path(file_okay=True, dir_okay=False),
              required=False,
              help='Filepath to write configurations to')
def config(debug, **kwargs):
    """Get job configuration"""
    set_debug_log_level(debug)
    cli_job.config(**translate_kwargs(kwargs))


@job.command(short_help='\tDisable job')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
def disable(debug, **kwargs):
    """Disable job"""
    set_debug_log_level(debug)
    cli_job.disable(**translate_kwargs(kwargs))


@job.command(short_help='\tEnable job')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
def enable(debug, **kwargs):
    """Enable job"""
    set_debug_log_level(debug)
    cli_job.enable(**translate_kwargs(kwargs))


@job.command(short_help='\tRename job')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
@click.argument('name', nargs=1, type=str, required=True)
def rename(debug, **kwargs):
    """Rename job"""
    set_debug_log_level(debug)
    cli_job.rename(**translate_kwargs(kwargs))


@job.command(short_help='\tDelete job')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
def delete(debug, **kwargs):
    """Delete job"""
    set_debug_log_level(debug)
    cli_job.delete(**translate_kwargs(kwargs))


@job.command(short_help='\tWipe job workspace')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=True)
def wipe(debug, **kwargs):
    """Wipe job workspace"""
    set_debug_log_level(debug)
    cli_job.wipe(**translate_kwargs(kwargs))


@job.command(short_help='\tStart monitor UI')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-s', '--sound', type=bool, required=False, is_flag=True, help='Enable sound effects')
def monitor(debug, **kwargs):
    """Start monitor UI"""
    set_debug_log_level(debug)
    cli_job.monitor(**translate_kwargs(kwargs))


@job.command(short_help='\tCreate a job')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.argument('folder', nargs=1, type=str, required=True)
@click.option('--config-file',
              default='',
              type=click.Path(file_okay=True, dir_okay=False),
              required=False,
              help='Path to local config file defining job')
@click.option('--config-is-json',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='The specified file is in JSON format')
def create(debug, **kwargs):
    """Create a job"""
    set_debug_log_level(debug)
    cli_job.create(**translate_kwargs(kwargs))


@job.command(short_help='\tGet job parameters')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@cli_decorators.list
def parameters(debug, **kwargs):
    """Get job's build parameters

    Build parameters are the parameters to be used when building a job.
    """
    set_debug_log_level(debug)
    cli_job.parameters(**translate_kwargs(kwargs))
